/* 
Copyright CNRS                                                           
contributor: Yann Guermeur 11/04/09

Yann.Guermeur@loria.fr

This software is a computer program whose purpose is to perform multi-category
discrimination with a quadratic loss multi-class support vector machine:
the M-SVM^2.

This software is governed by the CeCILL-B license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL-B
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the [CeCILL|CeCILL-B|CeCILL-C] license and that you accept its terms.
*/

/*--------------------------------------------------------------------------*/
/*  Name           : train_SVM.c                                            */
/*  Version        : 1.0                                                    */
/*  Creation       : 06/20/08                                               */
/*  Last update    : 05/14/10                                               */
/*  Subject        : Training algorithm of the M-SVM^2                      */
/*  Algo.          : Rosen's algorithm + decomposition method               */
/*  Author         : Yann Guermeur Yann.Guermeur@loria.fr                   */
/*--------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h> 
#include <string.h>
#include <math.h>
#include <ctype.h>
#include "algebre.h"

#define minimum(a,b) ((a)<=(b)?(a):(b))
#define maximum(a,b) ((a)>=(b)?(a):(b))

#define true 1
#define false 0
#define taille 81
#define very_small 1e-6
#define tiny 1e-12
#define nb_iter 1e9
#define step 1000

FILE *fs, *fc;

unsigned short xi[3];

long nature_kernel=0, nb_variables, nullity, optimum, index_smallest_u,
i, j, k, l, dim_input, nb_data, nb_constraints, iter=0, *y, row, column,
y_i, Q=0, chunk_size=0, *table_chunk, *out_of_chunk, ind_pattern, nb_sat,
nrand48(), random_num, active, nb_SV, descent, feasible=true, feasible_u;

char fichier_data[taille], fichier_alpha[taille], commande[taille],
fichier_fichcom[taille], fichier_alpha_0[taille];

double **X, **alpha, C=0.0, twoCinv=0.0, **gradient, partial, **K, **A0, **A0T,
**A0A0T, **A0A0T_1, **working_matrix1, **working_matrix2, smallest_u,
**eigenvectors, **eigenvectorsT, *eigenvalues, *mean, *variance, *st_dev,
**A0A0T_ref, **diagonal_matrix, **matrix_y_bar,
**new_diagonal_matrix, **vector_gradient, **vector_u, **vector_y_bar,
**vector_z_bar, norm, theta_max, **H_y_bar, theta_opt, sum_all_alpha=0.0,
gradientT_y_bar, y_barT_H_y_bar, val_obj_function=0.0, norm_y_bar,
initial_norm_y_bar, former_val_obj_function=0.0, *constraints;

/* Functions included in this program */

void caract_db();
void alloc_memory();
void read_data();
void standardize_data();
void read_alpha();
void compute_table_chunk();
void compute_K();
void compute_gradient();
void compute_A0();
void compute_vector_u();
void check_vector_u();
void compute_vector_y_bar();
void check_vector_y_bar();
void simplify_A0();
void compute_theta_max();
void compute_H_y_bar();
void compute_theta_opt();
void compute_new_alpha();
void check_feasible_sol();
void study_unfeasible_sol();
void write_vector(double **vector, char *fichier);
void Pause(char *message);
void compute_objective_function();
void compute_pseudo_inverse();

main(int argc, char *argv[])

{

system("clear");
srand48(getpid());
strcpy(fichier_fichcom, argv[1]);

caract_db();
read_data();
standardize_data();
/* standardize_data(); */
read_alpha();
check_feasible_sol();

for(iter=1; iter<=nb_iter; iter++)
  {

  if((iter%step) == 0)
    printf("\n\n*** Iteration: %6d\n", iter);
  compute_table_chunk();
  compute_K();
  compute_gradient();
  compute_A0();
  compute_vector_u();
  check_vector_u();

  if(feasible_u == true)
    {

    compute_vector_y_bar();
    check_vector_y_bar();

    if((nullity == true) && (optimum == false))
      while((nullity == true) && (feasible_u == true) && (optimum == false))
        {
        simplify_A0();
        compute_vector_u();
        check_vector_u();
	if((feasible_u == true) && (optimum == false))
          {
          compute_vector_y_bar();
          check_vector_y_bar();
          }
	}

    if((feasible_u == true) && (optimum == false) &&
       (nullity == false) && (descent == true))
      {
      compute_theta_max();
      compute_theta_opt();
      compute_new_alpha();
      check_feasible_sol();
      if(feasible == false)
        study_unfeasible_sol();
      }

    }

  if((iter%step) == 0)
    {
    write_vector(alpha, fichier_alpha);
/*
    compute_objective_function();
    Pause("");
*/
    }

  } 

}

void Pause(char *message)

/* Stop the execution after a display */
 
{

char c;
static int n = 0;

if(message!=NULL)
  printf("%s\n",message);

printf("\a\n=> New Line to proceed...\n");

n = read(0, &c, 1);

while(c != '\n')
  n = read(0, &c, 1);

return;

}

void caract_db()

/* Read the file specifying the learning problem */

{

if((fs=fopen(fichier_fichcom, "r"))==NULL)
  {
  printf("\nFile of parameters %s: cannot be open...\n", fichier_fichcom);
  exit(0);
  }

fscanf(fs, "%ld", &Q);
fscanf(fs, "%ld", &nature_kernel);
fscanf(fs, "%lf", &C);
twoCinv = 1.0 / (2.0 * C);
fscanf(fs, "%ld", &chunk_size);
fscanf(fs, "%s", fichier_data);
printf("\nThe file of the data is: %s", fichier_data);
fscanf(fs, "%s", fichier_alpha_0);
fscanf(fs, "%s", fichier_alpha);

fclose(fs);

sprintf(commande, "cp %s Save_alpha/.", fichier_alpha);

}

void alloc_memory()

/* Allocate the memory for some global variables */

{

nb_variables = chunk_size * (Q-1);
nb_constraints = nb_variables + (Q-1);

X = matrix(nb_data, dim_input);
mean = (double *) calloc(dim_input+1, sizeof(double));
variance = (double *) calloc(dim_input+1, sizeof(double));
st_dev = (double *) calloc(dim_input+1, sizeof(double));
y = (long *) calloc(nb_data+1, sizeof(long));
alpha = matrix(nb_data, Q);
constraints = (double *) calloc(Q+1, sizeof(double));
out_of_chunk = (long *) calloc(nb_data+1, sizeof(long));
gradient = matrix(chunk_size, Q);
H_y_bar = matrix(chunk_size, Q);
vector_gradient = matrix(nb_variables, 1);
table_chunk = (long *) calloc(chunk_size+1, sizeof(long));
K = matrix(chunk_size, nb_data);
A0 = matrix(nb_constraints, nb_variables);
A0T = matrix(nb_variables, nb_constraints);
A0A0T = matrix(nb_constraints, nb_constraints);
A0A0T_ref = matrix(nb_constraints, nb_constraints);
eigenvectors = matrix(nb_constraints, nb_constraints);
eigenvectorsT = matrix(nb_constraints, nb_constraints);
eigenvalues = (double *) calloc(nb_constraints+1, sizeof(double));
diagonal_matrix = matrix(nb_constraints, nb_constraints);
new_diagonal_matrix = matrix(nb_constraints, nb_constraints);
A0A0T_1 = matrix(nb_constraints, nb_constraints);
working_matrix1 = matrix(nb_constraints, nb_constraints);
working_matrix2 = matrix(nb_constraints, nb_constraints);
vector_u = matrix(nb_constraints, 1);
vector_y_bar = matrix(nb_variables, 1);
matrix_y_bar = matrix(chunk_size, Q);
vector_z_bar = matrix(nb_variables, 1);

}

void read_data()

/* Read the file containing the training set */

{

double value_y;

if((fs=fopen(fichier_data, "r"))==NULL)
  {
  printf("\nData file %s: cannot be open...\n", fichier_data);
  exit(0);
  }

fscanf(fs, "%ld", &nb_data);
fscanf(fs, "%ld", &dim_input);

alloc_memory();

for(i=1; i<=nb_data; i++)
  {
  for(j=1; j<=dim_input; j++)
    fscanf(fs, "%lf", &X[i][j]);
  fscanf(fs, "%lf", &value_y);
  y[i] = (long) value_y;
  }

fclose(fs);

printf("\nThe file of the data has been read...");

/*
display_mat(X, nb_data, dim_input, 6, 4);
Pause("");
*/

}

void standardize_data()

/* Standardize the data of the training set per predictor */

{

for(j=1; j<=dim_input; j++)
  {
  mean[j] = 0.0;
  variance[j] = 0.0;
  st_dev[j] = 0.0;
  }

for(i=1; i<=nb_data; i++)
  for(j=1; j<=dim_input; j++)
    mean[j] += X[i][j];

for(j=1; j<=dim_input; j++)
  {
  mean[j] /= (double) nb_data;
/*  printf("\nmean[%2d] = %e", j, mean[j]); */
  }
/* Pause(""); */

for(i=1; i<=nb_data; i++)
  for(j=1; j<=dim_input; j++)
    variance[j] += (X[i][j] - mean[j]) * (X[i][j] - mean[j]);

for(j=1; j<=dim_input; j++)
  {
  variance[j] /= (double) nb_data;
  st_dev[j] = sqrt(variance[j]);
/*  printf("\nst_dev[%2d] = %e", j, st_dev[j]); */
  }
/* Pause(""); */

for(i=1; i<=nb_data; i++)
  for(j=1; j<=dim_input; j++)
    {
    X[i][j] -= mean[j];
    if(st_dev[j] > very_small)
      X[i][j] /= st_dev[j];
    }

}

void read_alpha()

/* Read the initial values of the dual variables */

{

if((fs=fopen(fichier_alpha_0, "r"))==NULL)
  {
  printf(
  "\nFile of the initial values of the dual variables %s: cannot be open...\n",
  fichier_alpha_0);
  exit(0);
  }

for(i=1; i<=nb_data; i++)
  for(k=1; k<=Q; k++)
    fscanf(fs, "%lf", &alpha[i][k]);

fclose(fs);

printf(
"\nThe file of the initial values of the dual variables has been read...");

}

void compute_table_chunk()

/* Select a subset of the training set for the next step of optimization */

{

for(i=1; i<=nb_data; i++)
  out_of_chunk[i] = i;

nb_SV = 0;

for(i=1; i<=nb_data; i++)
  {
  active = 0;
  k = 1;
  while((active == 0) && (k <= Q))
    {
    if(alpha[i][k] > 0.0)
      {
      active = 1;
      nb_SV++;
      }
    else
      k++;
    }
  }

if((iter%step) == 0)
  printf("\nNumber of support vectors: %d", nb_SV);

for(i=1; i<=chunk_size; i++)
  {
  random_num = nrand48(xi);
  row = (random_num % (nb_data-i+1))+1;
  table_chunk[i] = out_of_chunk[row];
  for(j=row; j<=nb_data-i; j++)
    out_of_chunk[j] = out_of_chunk[j+1];
  }

}

void compute_K()

/* Compute the submatrix of the Gram matrix corresponding to the chunk */

{

for(i=1; i<=chunk_size; i++)
  {
  ind_pattern = table_chunk[i];
  for(j=1; j<=nb_data; j++)
    K[i][j] = ker(nature_kernel, X[ind_pattern], X[j], dim_input);
  K[i][ind_pattern] += twoCinv;
  }

}

void compute_gradient()

/* Compute the components of the gradient vector corresponding to the chunk */

{

double partial_sum_alpha = 0.0;

row = 0;

for(i=1; i<=chunk_size; i++)
  {
  y_i = y[table_chunk[i]];

  for(k=1; k<=Q; k++)
    {
    if(k != y_i)
      {
      gradient[i][k] = -1.0 / (Q - 1.0);
      for(j=1; j<=nb_data; j++)
        {
	partial_sum_alpha = 0.0;

        for(l=1; l<=Q; l++)
          if(l != y[j])
            partial_sum_alpha += alpha[j][l];

	partial = alpha[j][k] - (partial_sum_alpha / Q);
	gradient[i][k] += partial * K[i][j];
        }
      row++;
      vector_gradient[row][1] = gradient[i][k];
      }
    else
      gradient[i][k] = 0.0;
    }
  }

/*
printf("\nGradient (matrix form)\n");
display_mat(gradient, chunk_size, Q, 7, 4);

printf("\n\nGradient (vector)\n");
display_mat(vector_gradient, nb_variables, 1, 7, 4);
Pause("");
*/

}

void compute_A0()

/* Computation of the matrices A0 and A0T */

{

row=0;
column=0;

for(k=1; k<=nb_constraints; k++)
  for(l=1; l<=nb_variables; l++)
    A0[k][l] = 0.0;

for(i=1; i<=chunk_size; i++)
  {
  ind_pattern = table_chunk[i];
  y_i = y[ind_pattern];
  for(k=1; k<=Q; k++)
    {
    if(k != y_i)
      {
      column++;
      if(alpha[ind_pattern][k] == 0.0)
        {
        row++; 
        A0[row][column] = -1.0;
	}
      }
    }
  }

column=0;

for(i=1; i<=chunk_size; i++)
  {
  y_i = y[table_chunk[i]];

  for(l=1; l<=Q; l++)
    if(l != y_i)  
      {
      column++;
      for(k=1; k<Q; k++)
        A0[row+k][column] = (l != k) ? 1.0 : 1.0-Q;
      }

  }

nb_sat = row+Q-1;
trans_mat(A0, A0T, nb_sat, nb_variables);

}

void compute_vector_u()

/* Computation of the vector u satisfying A0A0T u = - A0 grad J(alpha)*/

{

mult_mat(A0, A0T, A0A0T, nb_sat, nb_variables, nb_sat);

/*
printf("\nMatrix A0A0T\n");
display_mat(A0A0T, nb_sat, nb_sat, 3, 0);
Pause("");
*/

compute_pseudo_inverse();

mult_mat(A0, vector_gradient, working_matrix1, nb_sat, nb_variables, 1);
mult_mat(A0A0T_1, working_matrix1, working_matrix2, nb_sat, nb_sat, 1);
scal_mat(-1.0, working_matrix2, vector_u, nb_sat, 1);

}

void check_vector_u()

/* Check the equation defining vector u */

{

mult_mat(A0A0T_ref, vector_u, working_matrix1, nb_sat, nb_sat, 1);
mult_mat(A0, vector_gradient, working_matrix2, nb_sat, nb_variables, 1);

for(k=1; k<=nb_sat; k++)
  working_matrix2[k][1] *= -1.0;

/*
printf("\nFirst vector\n");
display_mat(working_matrix1, nb_sat, 1, 7, 4);
Pause("");

printf("\nSecond vector\n");
display_mat(working_matrix2, nb_sat, 1, 7, 4);
Pause("");
*/

feasible_u = compare(working_matrix1, working_matrix2, nb_sat, 1, very_small);

optimum=true;
k=1;

while((k<=nb_sat-Q+1) && (optimum == true))
  {
  if(vector_u[k][1] < 0.0)
    optimum=false;
  else
    k++;
  }

}

void compute_vector_y_bar()

/* Compute the new direction of descent */

{

mult_mat(A0T, vector_u, vector_z_bar, nb_variables, nb_sat, 1);

/*
printf("\nVector bar{z}\n");
display_mat(vector_z_bar, nb_variables, 1, 7, 4);
*/

add_mat(vector_gradient, vector_z_bar, working_matrix1, nb_variables, 1);
scal_mat(-1.0, working_matrix1, vector_y_bar, nb_variables, 1);

/*
printf("\nVector bar{y}\n");
for(k=1; k<=nb_variables; k++)
  printf("\n%e", vector_y_bar[k][1]);
printf("\n");
*/

for(k=1; k<=nb_variables; k++)
  if(fabs(vector_y_bar[k][1]) < 0.01 * very_small)
    vector_y_bar[k][1] = 0.0;

/*
printf("\nVector bar{y}\n");
display_mat(vector_y_bar, nb_variables, 1, 7, 4);
*/

row=0;

for(i=1; i<=chunk_size; i++)
  for(k=1; k<=Q; k++)
    {
    if(k != y[table_chunk[i]])
      {
      row++;
      matrix_y_bar[i][k] = vector_y_bar[row][1];
      }
    else
      matrix_y_bar[i][k] = 0.0;
    }

/*
printf("\nMatrix bar{y}\n");
display_mat(matrix_y_bar, chunk_size, Q, 7, 4);
Pause("");
*/

}

void check_vector_y_bar()

/* Check the feasibility of the new direction of descent */

{

norm_y_bar=0.0;

for(k=1; k<=nb_variables; k++)
  norm_y_bar += vector_y_bar[k][1] * vector_y_bar[k][1];

norm_y_bar = sqrt(norm_y_bar);
initial_norm_y_bar = norm_y_bar;

if(norm_y_bar <= very_small)
  nullity=true;
else
  nullity=false;

if(nullity == false)
  {
  row=0;

  for(i=1; i<=chunk_size; i++)
    for(k=1; k<=Q; k++)
      {
      if(k != y[table_chunk[i]])
        {
        row++;
        vector_y_bar[row][1] /= norm_y_bar;
        matrix_y_bar[i][k] = vector_y_bar[row][1];
        }
      else
        matrix_y_bar[i][k] = 0.0;
      }

  descent=true;

  mult_mat(A0, vector_y_bar, working_matrix1, nb_sat, nb_variables, 1);
  norm=0.0;

  for(k=1; k<=nb_sat; k++)
    norm += working_matrix1[k][1] * working_matrix1[k][1];

  norm = sqrt(norm);

  if(norm > sqrt(nb_sat) * very_small)
    {
/*
printf("\nbar{y} does not belong to the null space of A0: ||A0 bar{y}||= %e\n",
       norm);
*/
    descent=false;
    }

  gradientT_y_bar=0.0;

  for(k=1; k<=nb_variables; k++)
    gradientT_y_bar += vector_y_bar[k][1] * vector_gradient[k][1];

  if(gradientT_y_bar > 0.0)
    {
    if(gradientT_y_bar > very_small)
      {
/*
  printf("\ngradient^T bar{y} = %e: bar{y} is not a direction of descent...\n",
         gradientT_y_bar);
*/
      }
    descent=false;
    }
  }

}

void simplify_A0()

/* Relax one of the constraints to be satisfied by the direction of descent */

{

smallest_u = 0.0;
index_smallest_u = 1;

for(k=1; k<=nb_sat-Q+1; k++)
  if(vector_u[k][1] < smallest_u)
    {
    smallest_u = vector_u[k][1];
    index_smallest_u = k;
    }

nb_sat--;

for(k=index_smallest_u; k<=nb_sat; k++)
  for(l=1; l<=nb_variables; l++)
    A0[k][l] = A0[k+1][l];

trans_mat(A0, A0T, nb_sat, nb_variables);

}

void compute_theta_max()

/* Computation of the largest possible value for theta */

{

double theta_partial;

theta_max = 0.0;

for(i=1; i<=chunk_size; i++)
  {
  ind_pattern = table_chunk[i];
  y_i = y[ind_pattern];

  for(k=1; k<=Q; k++)
    if((k != y_i) && (matrix_y_bar[i][k] < 0.0))
      {
      theta_partial = - alpha[ind_pattern][k] / matrix_y_bar[i][k];
      if((theta_partial < theta_max) || (theta_max == 0.0))
        theta_max = theta_partial;
      }
  }

/*
printf("\nTheta_max: %e\n", theta_max);
*/

}

void compute_H_y_bar()

/* Computation of the vector H bar{y} */

{

double partial_sum_y_bar = 0.0;

for(i=1; i<=chunk_size; i++)
  for(k=1; k<=Q; k++)
    {
    H_y_bar[i][k] = 0.0;
    if(k != y[table_chunk[i]])
      {
      for(j=1; j<=chunk_size; j++)
        {
        partial_sum_y_bar = 0.0;

        for(l=1; l<=Q; l++)
          if(l != y[table_chunk[j]])
            partial_sum_y_bar += matrix_y_bar[j][l];

        partial = matrix_y_bar[j][k] - (partial_sum_y_bar / Q);
        H_y_bar[i][k] += partial * K[i][table_chunk[j]];
	}
      }
    }

}

void compute_theta_opt()

/* Computation of the optimal value for theta */

{

y_barT_H_y_bar=0.0;
compute_H_y_bar();

for(i=1; i<=chunk_size; i++)
  for(k=1; k<=Q; k++)
    if(k != y[table_chunk[i]])
      y_barT_H_y_bar += matrix_y_bar[i][k] * H_y_bar[i][k];

if(y_barT_H_y_bar <= 0.0)
  {
  printf("\nThe value of the quadratic form is: %e\n", y_barT_H_y_bar);
  printf("\nMatrix bar{y}\n");
  display_mat(matrix_y_bar, chunk_size, Q, 7, 4);
  exit(0);
  }

if(y_barT_H_y_bar > tiny)
  {
  theta_opt = - gradientT_y_bar / y_barT_H_y_bar;
  if(theta_max > 0.0)
    theta_opt = minimum(theta_opt, theta_max);
  }
else
  {
  printf("\nbar{y}^T H bar{y} is too small: %e", y_barT_H_y_bar);
  theta_opt = 0.0;
  }

}

void compute_new_alpha()

/* Computation of the new value of the vector of dual variables */

{

for(i=1; i<=chunk_size; i++)
  {
  ind_pattern = table_chunk[i];
  y_i = y[ind_pattern];

  for(k=1; k<=Q; k++)
    if(k != y_i)
      {
      alpha[ind_pattern][k] += theta_opt * matrix_y_bar[i][k];
      if(alpha[ind_pattern][k] < 0.0)
        alpha[ind_pattern][k] = 0.0;
      }
  }

}

void check_feasible_sol()

/* Check the feasibility of the current solution alpha */

{

sum_all_alpha = 0.0;

for(i=1; i<=nb_data; i++)
  for(k=1; k<=Q; k++)
    if(((k == y[i]) && (alpha[i][k] != 0.0)) || (alpha[i][k] < 0.0))
      {
      printf("\nNo feasible solution: alpha[%d][%d] = %e\n\n",
             i, k, alpha[i][k]);
      exit(0);
      }
    else
      sum_all_alpha += alpha[i][k];

for(k=1; k<=Q; k++)
  constraints[k] = sum_all_alpha / Q;

for(i=1; i<=nb_data; i++)
  for(k=1; k<=Q; k++)
    constraints[k] -= alpha[i][k];

norm = 0.0;

for(k=1; k<=Q; k++)
  norm += constraints[k] * constraints[k];

norm = sqrt(norm);

if(norm >= very_small)
  {
  feasible = false;
  printf("\n\nLarge deviation of the equality constraints...\n\n");
  }
else
  if(iter == 0)
    printf("\nVector alpha is a feasible solution...\n");

}

void study_unfeasible_sol()

/* Characterize the achieving of an unfeasible solution */

{

for(k=1; k<=Q; k++)
  printf("%s%e\n", constraints[k] < 0.0 ? " " : "  ", constraints[k]);

printf("\nMatrix A0\n");
display_mat(A0, nb_sat, nb_variables, 7, 4);

if(nullity == true)
  printf("\nbar{y} is the null vector");
else
  printf("\nbar{y} is not the null vector");
printf("\nNorm of bar{y}: %e\n", initial_norm_y_bar);

printf("\nVector u\n");
display_mat(vector_u, nb_sat, 1, 7, 4);

printf("\nVector bar{y}\n");
display_mat(matrix_y_bar, chunk_size, Q, 7, 4);

mult_mat(A0, vector_y_bar, working_matrix1, nb_sat, nb_variables, 1);
printf("\nA0 bar{y}\n");
display_mat(working_matrix1, nb_sat, 1, 9, 6);

exit(0);

}

void write_vector(double **vector, char *fichier)

/* Write a vector/matrix in a file (used for the vector alpha) */

{

if((fc=fopen(fichier, "w"))==NULL)
  {
  printf("\nThe file %s cannot be open...\n", fichier);
  exit(0);
  }

for(i=1; i<=nb_data; i++)
  for(k=1; k<=Q; k++)
    fprintf(fc, "%15.10f %c", vector[i][k], (k==Q) ? '\n': ' ');

fclose(fc);

system(commande);

}

void compute_objective_function()

/* Computation of the value J(alpha) of the objective function */

{

double Q_term=0.0, L_term=0.0, partial_1, partial_2, kernel;

for(i=1; i<=nb_data; i++)
  for(j=1; j<=nb_data; j++)
    {
    partial_1 = 0.0;
    for(k=1; k<=Q; k++)
      for(l=1; l<=Q; l++)
        {
        partial_2 = -1.0;
        if(k == l) 
          partial_2 += Q;
	partial_1 += partial_2 * alpha[i][k] * alpha[j][l];
        }
    if(partial_1 != 0.0)
      {
      kernel = ker(nature_kernel, X[i], X[j], dim_input); 
      if(i == j)
        kernel += twoCinv;
      Q_term += partial_1 * kernel;
      }
    }

Q_term /= Q;
L_term = sum_all_alpha / (Q - 1.0);
val_obj_function = -0.5 * Q_term + L_term;

printf("\n       Objective function: %e", val_obj_function);

if(val_obj_function < former_val_obj_function)
  if(former_val_obj_function - val_obj_function > very_small)
    {
    printf("\nThe value of the objective function has decreased...\n");
    printf("%e < %e\n\n", val_obj_function, former_val_obj_function);
    exit(0);
    }
else
  former_val_obj_function = val_obj_function;

}

void compute_pseudo_inverse()

/* Computation of the pseudo_inverse of A0A0T */

{

copy(A0A0T, A0A0T_ref, nb_sat, nb_sat);
diagonalize(A0A0T, nb_sat, eigenvalues);
copy(A0A0T, eigenvectors, nb_sat, nb_sat);
trans_mat(eigenvectors, eigenvectorsT, nb_sat, nb_sat);

/*
for(k=1; k<=nb_sat; k++)
  diagonal_matrix[k][k] = eigenvalues[k];

printf("\nDiagonal matrix\n");
display_mat(diagonal_matrix, nb_sat, nb_sat, 7, 3);
Pause("");

mult_mat(A0A0T_ref, eigenvectors, working_matrix1, nb_sat, nb_sat, nb_sat);
mult_mat(eigenvectorsT, working_matrix1, working_matrix2, 
nb_sat, nb_sat, nb_sat);

printf("\nDiagonal matrix rebuilt\n");
display_mat(working_matrix2, nb_sat, nb_sat, 7, 3);
Pause("");

if(compare(working_matrix2, diagonal_matrix, nb_sat, nb_sat, very_small) 
   == false)
  {
  printf("\nProblem in the computation of the eigenvalues...\n");
  exit(0);
  }
else
  printf("\nNo problem in the computation of the eigenvalues...");

mult_mat(diagonal_matrix, eigenvectorsT, working_matrix1, 
nb_sat, nb_sat, nb_sat);

mult_mat(eigenvectors, working_matrix1, working_matrix2,
nb_sat, nb_sat, nb_sat);

printf("\nMatrix A0A0T rebuilt\n");
display_mat(working_matrix2, nb_sat, nb_sat, 7, 3);
Pause("");

if(compare(working_matrix2, A0A0T_ref, nb_sat, nb_sat, very_small) == false)
  {
  printf("\nLarge deviation resulting from the diagonalization...\n");
  exit(0);
  }
else
  printf("\nNo deviation resulting from the diagonalization...");
*/

for(k=1; k<=nb_sat; k++)
  if(fabs(eigenvalues[k]) > very_small)
    new_diagonal_matrix[k][k] = 1.0 / eigenvalues[k];
  else
    new_diagonal_matrix[k][k] = 0.0;

mult_mat(new_diagonal_matrix, eigenvectorsT, working_matrix1,
	 nb_sat, nb_sat, nb_sat);

mult_mat(eigenvectors, working_matrix1, A0A0T_1,
         nb_sat, nb_sat, nb_sat);

/*
mult_mat(A0A0T_ref, A0A0T_1, working_matrix1, nb_sat, nb_sat, nb_sat);
mult_mat(working_matrix1, A0A0T_ref, A0A0T, nb_sat, nb_sat, nb_sat);

printf("\nMatrix A0A0T rebuilt\n");
display_mat(A0A0T, nb_sat, nb_sat, 7, 3);
Pause("");

if(compare(A0A0T, A0A0T_ref, nb_sat, nb_sat, very_small) == false)
  {
  printf("\nProblem in the computation of the pseudo-inverse...\n");
  exit(0);
  }
else
  printf("\nThe pseudo-inverse has been computed...");
*/

}
